@@ -38,18 +38,41 @@ except ImportError:
     from httplib import BadStatusLine  # type: ignore
     from urllib2 import HTTPError, urlopen  # type: ignore
 
-from collections import Mapping, Sequence
-
 try:
-    from typing import Any, Iterable, Iterator, List, Set, Tuple, Union
+    from typing import (Any, Iterable, Iterator, List, Optional, Set,
+                        Tuple, Union, no_type_check, overload, Mapping,
+                        Sequence)
     Any  # pyflakes
     Iterable  # pyflakes
     Iterator  # pyflakes
     List  # pyflakes
+    Optional  # pyflakes
     Set  # pyflakes
     Tuple  # pyflakes
     Union  # pyflakes
+    overload  # pyflakes
+    Sequence  # pyflakes
 except ImportError:
+    import collections
+
+    class GenericWrapper(object):
+        """Takes a non-generic type and adds __getitem__"""
+        def __init__(self, value):
+            # type: (type) -> None
+            self.value = value
+
+        def __getitem__(self, key):
+            # type: (type) -> type
+            return self.value
+
+    List = GenericWrapper(list)  # type: ignore
+    Mapping = GenericWrapper(collections.Mapping)  # type: ignore
+    Sequence = GenericWrapper(collections.Sequence)  # type: ignore
+    Any = None
+
+    def no_type_check(arg):
+        # type: (Any) -> Any
+        return arg
     pass
 
 import apt_pkg
@@ -71,12 +94,12 @@ if sys.version_info.major >= 3:
     unicode = str
 
 
-def _file_is_same(path, size, md5):
-    # type: (str, int, str) -> bool
+def _file_is_same(path, size, hashes):
+    # type: (str, int, apt_pkg.HashStringList) -> bool
     """Return ``True`` if the file is the same."""
     if os.path.exists(path) and os.path.getsize(path) == size:
         with open(path) as fobj:
-            return apt_pkg.md5sum(fobj) == md5
+            return apt_pkg.Hashes(fobj).hashes == hashes
     return False
 
 
@@ -84,6 +107,10 @@ class FetchError(Exception):
     """Raised when a file could not be fetched."""
 
 
+class UntrustedError(FetchError):
+    """Raised when a file did not have a trusted hash."""
+
+
 class BaseDependency(object):
     """A single dependency."""
 
@@ -97,6 +124,7 @@ class BaseDependency(object):
         """
 
         def __eq__(self, other):
+            # type: (object) -> bool
             if str.__eq__(self, other):
                 return True
             elif str.__eq__(self, '<'):
@@ -109,6 +137,7 @@ class BaseDependency(object):
                 return False
 
         def __ne__(self, other):
+            # type: (object) -> bool
             return not self.__eq__(other)
 
     def __init__(self, version, dep):
@@ -172,9 +201,9 @@ class BaseDependency(object):
         .. versionadded:: 1.0.0
         """
         tvers = []
-        _tvers = self._dep.all_targets()  # [apt_pkg.Version, ...]
-        for _tver in _tvers:  # apt_pkg.Version
-            _pkg = _tver.parent_pkg  # apt_pkg.Package
+        _tvers = self._dep.all_targets()  # type: List[apt_pkg.Version]
+        for _tver in _tvers:  # type: apt_pkg.Version
+            _pkg = _tver.parent_pkg  # type: apt_pkg.Package
             cache = self._version.package._pcache  # apt.cache.Cache
             pkg = cache._rawpkg_to_pkg(_pkg)  # apt.package.Package
             tver = Version(pkg, _tver)  # apt.package.Version
@@ -231,7 +260,7 @@ class BaseDependency(object):
         return self._dep.dep_type_untranslated == 'PreDepends'
 
 
-class Dependency(list):
+class Dependency(List[BaseDependency]):
     """Represent an Or-group of dependencies.
 
     Attributes defined here:
@@ -242,8 +271,8 @@ class Dependency(list):
     """
 
     def __init__(self, version, base_deps, rawtype):
-        # type: (Version, apt_pkg.Dependency, str) -> None
-        super(Dependency, self).__init__(base_deps)  # type: ignore
+        # type: (Version, List[BaseDependency], str) -> None
+        super(Dependency, self).__init__(base_deps)
         self._version = version  # apt.package.Version
         self._rawtype = rawtype
 
@@ -351,7 +380,7 @@ class Origin(object):
                                             self.site, self.trusted)
 
 
-class Record(Mapping):
+class Record(Mapping[Any, Any]):
     """Record in a Packages file
 
     Represent a record as stored in a Packages file. You can use this like
@@ -376,7 +405,7 @@ class Record(Mapping):
         self._rec = apt_pkg.TagSection(record_str)
 
     def __hash__(self):
-        # type: () -> Any
+        # type: () -> int
         return hash(self._rec)
 
     def __str__(self):
@@ -393,10 +422,10 @@ class Record(Mapping):
 
     def __iter__(self):
         # type: () -> Iterator[str]
-        return iter(self._rec.keys())
+        return iter(self._rec.keys())   # type: ignore
 
     def iteritems(self):
-        # type: () -> Iterable
+        # type: () -> Iterable[Tuple[object, str]]
         """An iterator over the (key, value) items of the record."""
         for key in self._rec.keys():
             yield key, self._rec[key]
@@ -432,9 +461,10 @@ class Version(object):
         # type: (Package, apt_pkg.Version) -> None
         self.package = package
         self._cand = cand
+        self.package._pcache._weakversions.add(self)
 
     def _cmp(self, other):
-        # FIXME: add type hint
+        # type: (Any) -> Union[int, NotImplemented]
         """Compares against another apt.Version object or a version string.
 
         This method behaves like Python 2's cmp builtin and returns an integer
@@ -484,14 +514,14 @@ class Version(object):
         return self._cmp(other) < 0
 
     def __ne__(self, other):
-        # type: (object) -> bool
+        # type: (object) -> Union[bool, NotImplemented]
         try:
             return self._cmp(other) != 0
         except TypeError:
             return NotImplemented
 
     def __hash__(self):
-        # type: () -> Any
+        # type: () -> int
         return self._cand.hash
 
     def __str__(self):
@@ -507,17 +537,20 @@ class Version(object):
     def _records(self):
         # type: () -> apt_pkg.PackageRecords
         """Internal helper that moves the Records to the right position."""
-        if self.package._pcache._records.lookup(self._cand.file_list[0]):
-            return self.package._pcache._records
-        return None
+        # If changing lookup, change fetch_binary() as well
+        if not self.package._pcache._records.lookup(self._cand.file_list[0]):
+            raise LookupError("Could not lookup record")
+
+        return self.package._pcache._records
 
     @property
     def _translated_records(self):
-        # type: () -> apt_pkg.PackageRecords
+        # type: () -> Optional[apt_pkg.PackageRecords]
         """Internal helper to get the translated description."""
         desc_iter = self._cand.translated_description
-        self.package._pcache._records.lookup(desc_iter.file_list.pop(0))
-        return self.package._pcache._records
+        if self.package._pcache._records.lookup(desc_iter.file_list.pop(0)):
+            return self.package._pcache._records
+        return None
 
     @property
     def installed_size(self):
@@ -567,9 +600,10 @@ class Version(object):
 
     @property
     def summary(self):
-        # type: () -> str
+        # type: () -> Optional[str]
         """Return the short description (one line summary)."""
-        return self._translated_records.short_desc
+        records = self._translated_records
+        return records.short_desc if records is not None else None
 
     @property
     def raw_description(self):
@@ -594,7 +628,13 @@ class Version(object):
         for more information.
         """
         desc = ''
-        dsc = self._translated_records.long_desc
+        records = self._translated_records
+        dsc = records.long_desc if records is not None else None
+
+        if not dsc:
+            return _("Missing description for '%s'."
+                     "Please report.") % (self.package.name)
+
         try:
             if not isinstance(dsc, unicode):
                 # Only convert where needed (i.e. Python 2.X)
@@ -661,11 +701,7 @@ class Version(object):
         """Return the internal policy priority as a number.
            See apt_preferences(5) for more information about what it means.
         """
-        priority = 0
-        policy = self.package._pcache._depcache.policy
-        for (packagefile, _unused) in self._cand.file_list:
-            priority = max(priority, policy.get_priority(packagefile))
-        return priority
+        return self.package._pcache._depcache.policy.get_priority(self._cand)
 
     @property
     def record(self):
@@ -678,7 +714,7 @@ class Version(object):
         return Record(self._records.record)
 
     def get_dependencies(self, *types):
-        # FIXME: add type hints
+        # type: (str) -> List[Dependency]
         """Return a list of Dependency objects for the given types.
 
         Multiple types can be specified. Possible types are:
@@ -808,7 +844,7 @@ class Version(object):
 
     @property
     def uri(self):
-        # type: () -> str
+        # type: () -> Optional[str]
         """Return a single URI for the binary.
 
         .. versionadded:: 0.7.10
@@ -818,8 +854,9 @@ class Version(object):
         except StopIteration:
             return None
 
-    def fetch_binary(self, destdir='', progress=None):
-        # type: (str, AcquireProgress) -> str
+    def fetch_binary(self, destdir='', progress=None,
+                     allow_unauthenticated=None):
+        # type: (str, Optional[AcquireProgress], Optional[bool]) -> str
         """Fetch the binary version of the package.
 
         The parameter *destdir* specifies the directory where the package will
@@ -829,15 +866,39 @@ class Version(object):
         object. If not specified or None, apt.progress.text.AcquireProgress()
         is used.
 
+        The keyword-only parameter *allow_unauthenticated* specifies whether
+        to allow unauthenticated downloads. If not specified, it defaults to
+        the configuration option `APT::Get::AllowUnauthenticated`.
+
         .. versionadded:: 0.7.10
         """
+        if allow_unauthenticated is None:
+            allow_unauthenticated = apt_pkg.config.find_b("APT::Get::"
+                                        "AllowUnauthenticated", False)
         base = os.path.basename(self._records.filename)
         destfile = os.path.join(destdir, base)
-        if _file_is_same(destfile, self.size, self._records.md5_hash):
+        if _file_is_same(destfile, self.size, self._records.hashes):
             logging.debug('Ignoring already existing file: %s' % destfile)
             return os.path.abspath(destfile)
+
+        # Verify that the index is actually trusted
+        pfile, offset = self._cand.file_list[0]
+        index = self.package._pcache._list.find_index(pfile)
+
+        if not (allow_unauthenticated or (index and index.is_trusted)):
+            raise UntrustedError("Could not fetch %s %s source package: "
+                                 "Source %r is not trusted" %
+                                 (self.package.name, self.version,
+                                  getattr(index, "describe", "<unkown>")))
+        if not self.uri:
+            raise ValueError("No URI for this binary.")
+        hashes = self._records.hashes
+        if not (allow_unauthenticated or hashes.usable):
+            raise UntrustedError("The item %r could not be fetched: "
+                                     "No trusted hash found." %
+                                     destfile)
         acq = apt_pkg.Acquire(progress or apt.progress.text.AcquireProgress())
-        acqfile = apt_pkg.AcquireFile(acq, self.uri, self._records.md5_hash,
+        acqfile = apt_pkg.AcquireFile(acq, self.uri, hashes,
                                       self.size, base, destfile=destfile)
         acq.run()
 
@@ -847,8 +908,9 @@ class Version(object):
 
         return os.path.abspath(destfile)
 
-    def fetch_source(self, destdir="", progress=None, unpack=True):
-        # type: (str, AcquireProgress, bool) -> str
+    def fetch_source(self, destdir="", progress=None, unpack=True,
+                     allow_unauthenticated=None):
+        # type: (str, Optional[AcquireProgress], bool, Optional[bool]) -> str
         """Get the source code of a package.
 
         The parameter *destdir* specifies the directory where the source will
@@ -863,7 +925,15 @@ class Version(object):
 
         If *unpack* is ``True``, the path to the extracted directory is
         returned. Otherwise, the path to the .dsc file is returned.
+
+        The keyword-only parameter *allow_unauthenticated* specifies whether
+        to allow unauthenticated downloads. If not specified, it defaults to
+        the configuration option `APT::Get::AllowUnauthenticated`.
         """
+        if allow_unauthenticated is None:
+            allow_unauthenticated = apt_pkg.config.find_b("APT::Get::"
+                                        "AllowUnauthenticated", False)
+
         src = apt_pkg.SourceRecords()
         acq = apt_pkg.Acquire(progress or apt.progress.text.AcquireProgress())
 
@@ -878,18 +948,33 @@ class Version(object):
         if not source_lookup:
             raise ValueError("No source for %r" % self)
         files = list()
-        for md5, size, path, type_ in src.files:
-            base = os.path.basename(path)
+
+        if not (allow_unauthenticated or src.index.is_trusted):
+            raise UntrustedError("Could not fetch %s %s source package: "
+                                 "Source %r is not trusted" %
+                                 (self.package.name, self.version,
+                                  src.index.describe))
+        for fil in src.files:
+            base = os.path.basename(fil.path)
             destfile = os.path.join(destdir, base)
-            if type_ == 'dsc':
+            if fil.type == 'dsc':
                 dsc = destfile
-            if _file_is_same(destfile, size, md5):
+            if _file_is_same(destfile, fil.size, fil.hashes):
                 logging.debug('Ignoring already existing file: %s' % destfile)
                 continue
-            files.append(apt_pkg.AcquireFile(acq, src.index.archive_uri(path),
-                         md5, size, base, destfile=destfile))
+
+            if not (allow_unauthenticated or fil.hashes.usable):
+                raise UntrustedError("The item %r could not be fetched: "
+                                         "No trusted hash found." %
+                                         destfile)
+            files.append(apt_pkg.AcquireFile(acq,
+                            src.index.archive_uri(fil.path),
+                            fil.hashes, fil.size, base, destfile=destfile))
         acq.run()
 
+        if dsc is None:
+            raise ValueError("No source for %r" % self)
+
         for item in acq.items:
             if item.status != item.STAT_DONE:
                 raise FetchError("The item %r could not be fetched: %s" %
@@ -904,7 +989,7 @@ class Version(object):
             return os.path.abspath(dsc)
 
 
-class VersionList(Sequence):
+class VersionList(Sequence[Version]):
     """Provide a mapping & sequence interface to all versions of a package.
 
     This class can be used like a dictionary, where version strings are the
@@ -925,19 +1010,21 @@ class VersionList(Sequence):
     """
 
     def __init__(self, package, slice_=None):
-        # type: (Package, Any) -> None
+        # type: (Package, Optional[slice]) -> None
         self._package = package  # apt.package.Package()
         self._versions = package._pkg.version_list  # [apt_pkg.Version(), ...]
         if slice_:
             self._versions = self._versions[slice_]
 
     def __getitem__(self, item):
-        # FIXME: add type hints
+        # type: (Union[int, slice, str]) -> Any
+        # FIXME: Should not be returning Any, should have overloads; but
+        # pyflakes complains
         if isinstance(item, slice):
             return self.__class__(self._package, item)
         try:
             # Sequence interface, item is an integer
-            return Version(self._package, self._versions[item])
+            return Version(self._package, self._versions[item])  # type: ignore
         except TypeError:
             # Dictionary interface item is a string.
             for ver in self._versions:
@@ -984,10 +1071,10 @@ class VersionList(Sequence):
         return [ver.ver_str for ver in self._versions]
 
     def get(self, key, default=None):
-        # type: (str, Version) -> Version
+        # type: (str, Optional[Version]) -> Optional[Version]
         """Return the key or the default."""
         try:
-            return self[key]
+            return self[key]  # type: ignore  # FIXME: should be deterined automatically # noqa
         except LookupError:
             return default
 
@@ -1020,8 +1107,9 @@ class Package(object):
         # type: (Package) -> bool
         return self.name < other.name
 
-    def __get_candidate(self):
-        # type: () -> Version
+    @property
+    def candidate(self):
+        # type: () -> Optional[Version]
         """Return the candidate version of the package.
 
         This property is writeable to allow you to set the candidate version
@@ -1033,18 +1121,17 @@ class Package(object):
             return Version(self, cand)
         return None
 
-    def __set_candidate(self, version):
+    @candidate.setter
+    def candidate(self, version):
         # type: (Version) -> None
         """Set the candidate version of the package."""
         self._pcache.cache_pre_change()
         self._pcache._depcache.set_candidate_ver(self._pkg, version._cand)
         self._pcache.cache_post_change()
 
-    candidate = property(__get_candidate, __set_candidate)
-
     @property
     def installed(self):
-        # type: () -> Version
+        # type: () -> Optional[Version]
         """Return the currently installed version of the package.
 
         .. versionadded:: 0.7.9
@@ -1096,13 +1183,6 @@ class Package(object):
         This can be used eg. to store additional information about the pkg."""
         return self._pkg.id
 
-    def __hash__(self):
-        # type: () -> int
-        """Return the hash of the object.
-
-        This returns the same value as ID, which is unique."""
-        return self._pkg.id
-
     @property
     def essential(self):
         # type: () -> bool
@@ -1125,12 +1205,6 @@ class Package(object):
         """
         return self._pkg.architecture
 
-    @property
-    def section(self):
-        # type: () -> str
-        """Return the section of the package."""
-        return self._pkg.section
-
     # depcache states
 
     @property
@@ -1220,7 +1294,7 @@ class Package(object):
         return []
 
     def get_changelog(self, uri=None, cancel_lock=None):
-        # type: (str, threading.Event) -> str
+        # type: (Optional[str], Optional[threading.Event]) -> str
         """
         Download the changelog of the package and return it as unicode
         string.
@@ -1241,9 +1315,10 @@ class Package(object):
         if self._changelog != u"":
             return self._changelog
 
+        if not self.candidate:
+            return _("The list of changes is not available")
+
         if uri is None:
-            if not self.candidate:
-                pass
             if self.candidate.origins[0].origin == "Debian":
                 uri = "http://packages.debian.org/changelogs/pool" \
                       "/%(src_section)s/%(prefix)s/%(src_pkg)s" \
@@ -1329,7 +1404,7 @@ class Package(object):
                 changelog_file = urlopen(uri)
                 # do only get the lines that are new
                 changelog = u""
-                regexp = "^%s \((.*)\)(.*)$" % (re.escape(src_pkg))
+                regexp = "^%s \\((.*)\\)(.*)$" % (re.escape(src_pkg))
                 while True:
                     # Check if the download was canceled
                     if cancel_lock and cancel_lock.is_set():
@@ -1446,7 +1521,6 @@ class Package(object):
             fix.clear(self._pkg)
             fix.protect(self._pkg)
             fix.remove(self._pkg)
-            fix.install_protect()
             fix.resolve()
         self._pcache.cache_post_change()
 
@@ -1510,6 +1584,7 @@ class Package(object):
         self._pcache._depcache.commit(fprogress, iprogress)
 
 
+@no_type_check
 def _test():
     """Self-test."""
     print("Self-test for the Package modul")
