@@ -109,7 +109,7 @@ class SourceEntry(object):
         """ equal operator for two sources.list entries """
         return (self.disabled == other.disabled and
                 self.type == other.type and
-                self.uri == other.uri and
+                self.uri.rstrip('/') == other.uri.rstrip('/') and
                 self.dist == other.dist and
                 self.comps == other.comps)
 
@@ -153,8 +153,8 @@ class SourceEntry(object):
     def parse(self, line):
         """ parse a given sources.list (textual) line and break it up
             into the field we have """
-        line = self.line.strip()
-        #print line
+        self.line = line
+        line = line.strip()
         # check if the source is enabled/disabled
         if line == "" or line == "#":  # empty line
             self.invalid = True
@@ -310,12 +310,16 @@ class SourcesList(object):
         reuse them as far as possible
         """
 
+        type = type.strip()
+        disabled = type.startswith("#")
+        if disabled:
+            type = type[1:].lstrip()
         architectures = set(architectures)
         # create a working copy of the component list so that
         # we can modify it later
         comps = orig_comps[:]
         sources = self.__find(lambda s: set(s.architectures) == architectures,
-                              disabled=False, invalid=False, type=type,
+                              disabled=disabled, invalid=False, type=type,
                               uri=uri, dist=dist)
         # check if we have this source already in the sources.list
         for source in sources:
@@ -330,29 +334,38 @@ class SourcesList(object):
         sources = self.__find(lambda s: set(s.architectures) == architectures,
                               invalid=False, type=type, uri=uri, dist=dist)
         for source in sources:
-            # if there is a repo with the same (type, uri, dist) just add the
-            # components
-            if source.disabled and set(source.comps) == set(comps):
-                source.disabled = False
-                return source
-            elif not source.disabled:
-                source.comps = uniq(source.comps + comps)
+            if source.disabled == disabled:
+                # if there is a repo with the same (disabled, type, uri, dist)
+                # just add the components
+                if set(source.comps) != set(comps):
+                    source.comps = uniq(source.comps + comps)
                 return source
+            elif source.disabled and not disabled:
+                # enable any matching (type, uri, dist), but disabled repo
+                if set(source.comps) == set(comps):
+                    source.disabled = False
+                    return source
         # there isn't any matching source, so create a new line and parse it
-        line = type
-        if architectures:
-            line += " [arch=%s]" % ",".join(architectures)
-        line += " %s %s" % (uri, dist)
-        for c in comps:
-            line = line + " " + c
-        if comment != "":
-            line = "%s #%s\n" % (line, comment)
-        line = line + "\n"
+        parts = [
+            "#" if disabled else "",
+            type,
+            ("[arch=%s]" % ",".join(architectures)) if architectures else "",
+            uri,
+            dist,
+        ]
+        parts.extend(comps)
+        if comment:
+            parts.append("#" + comment)
+        line = " ".join(part for part in parts if part) + "\n"
+
         new_entry = SourceEntry(line)
         if file is not None:
             new_entry.file = file
         self.matcher.match(new_entry)
-        self.list.insert(pos, new_entry)
+        if pos < 0:
+            self.list.append(new_entry)
+        else:
+            self.list.insert(pos, new_entry)
         return new_entry
 
     def remove(self, source_entry):
